''' MariaDB status flags

These flas describe the current status of the database server.
'''

IN_TRANS = 1
AUTOCOMMIT = 2
MORE_RESULTS_EXIST = 8
QUERY_NO_GOOD_INDEX_USED =16
QUERY_NO_INDEX_USED = 32
CURSOR_EXISTS = 64
LAST_ROW_SENT = 128
DB_DROPPED = 256 
NO_BACKSLASH_ESCAPES = 512
METADATA_CHANGED = 1024
QUERY_WAS_SLOW = 2048
PS_OUT_PARAMS = 4096
IN_TRANS_READONLY = 8192
SESSION_STATE_CHANGED = 16384
ANSI_QUOTES = 32768



from mariadb.constants import FIELD_TYPE, FIELD_FLAG
import time, datetime

apilevel= '2.0'

paramstyle = 'qmark'

threadsafety = True

class DbApiType(frozenset):
    """
    Immutable set for type checking

    By default the following sets are defined:

    - BINARY: for binary field types
    - NUMBER: for numeric field types
    - STRING: for character based (string) field types
    - DATE: for date field type(s)
    - DATETIME: for datetime and timestamp field type(s)
    - TIME: for time field type(s)
    - TIMESTAMP: for datetime and timestamp field type(s)


    Example:
    >>> FIELD_TYPE.GEOMETRY == mariadb.BINARY
    True
    >>> FIELD_TYPE.FLOAT == mariadb.BINARY
    False
    """

    def __eq__(self, field_type):
        if (isinstance(field_type, DbApiType)):
            return not self.difference(field_type)
        return field_type in self

BINARY = DbApiType([FIELD_TYPE.GEOMETRY,
                    FIELD_TYPE.LONG_BLOB,
                    FIELD_TYPE.MEDIUM_BLOB,
                    FIELD_TYPE.TINY_BLOB,
                    FIELD_TYPE.BLOB])

STRING = DbApiType([FIELD_TYPE.ENUM,
                    FIELD_TYPE.JSON,
                    FIELD_TYPE.STRING,
                    FIELD_TYPE.VARCHAR,
                    FIELD_TYPE.VAR_STRING])

NUMBER = DbApiType([FIELD_TYPE.DECIMAL,
                    FIELD_TYPE.DOUBLE,
                    FIELD_TYPE.FLOAT,
                    FIELD_TYPE.INT24,
                    FIELD_TYPE.LONG,
                    FIELD_TYPE.LONGLONG,
                    FIELD_TYPE.NEWDECIMAL,
                    FIELD_TYPE.SHORT,
                    FIELD_TYPE.TINY,
                    FIELD_TYPE.YEAR])

DATE = DbApiType([FIELD_TYPE.DATE])
TIME = DbApiType([FIELD_TYPE.TIME])
DATETIME = TIMESTAMP = DbApiType([FIELD_TYPE.DATETIME,
                                          FIELD_TYPE.TIMESTAMP])
ROWID = DbApiType()

def Binary(object):
    """Constructs an object capable of holding a binary value."""
    return bytes(object)

def Date(year, month, day):
    """Constructs an object holding a date value."""
    return datetime.date(year, month, day)

def Time(hour, minute, second):
    """Constructs an object holding a time value."""
    return datetime.time(hour, minute, second)

def Timestamp(year, month, day, hour, minute, second):
    """Constructs an object holding a datetime value."""
    return datetime.datetime(year, month, day, hour, minute, second)

def DateFromTicks(ticks):
    """Constructs an object holding a date value from the given ticks value
       (number of seconds since the epoch).
       For more information see the documentation of the standard Python
       time module."""
    return Date(*time.localtime(ticks)[:3])

def TimeFromTicks(ticks):
    """Constructs an object holding a time value from the given ticks value
       (number of seconds since the epoch).
       For more information see the documentation of the standard Python
       time module."""
    return Time(*time.localtime(ticks)[3:6])

def TimestampFromTicks(ticks):
    """Constructs an object holding a datetime value from the given ticks value
       (number of seconds since the epoch).
       For more information see the documentation of the standard Python
       time module."""
    return datetime.datetime(*time.localtime(ticks)[:6])



from mariadb.constants import FIELD_TYPE, FIELD_FLAG

field_types= {FIELD_TYPE.DECIMAL : "DECIMAL",
              FIELD_TYPE.TINY : "TINY",
              FIELD_TYPE.SHORT : "SHORT",
              FIELD_TYPE.LONG  : "LONG",
              FIELD_TYPE.FLOAT : "FLOAT",
              FIELD_TYPE.DOUBLE : "DOUBLE",
              FIELD_TYPE.NULL : "NULL",
              FIELD_TYPE.TIMESTAMP : "TIMESTAMP",
              FIELD_TYPE.LONGLONG : "LONGLONG",
              FIELD_TYPE.INT24 : "INT24",
              FIELD_TYPE.DATE : "DATE",
              FIELD_TYPE.TIME : "TIME",
              FIELD_TYPE.DATETIME : "DATETIME",
              FIELD_TYPE.YEAR : "YEAR",
              FIELD_TYPE.NEWDATE : "NEWDATE",
              FIELD_TYPE.VARCHAR : "VARCHAR",
              FIELD_TYPE.BIT : "BIT",
              FIELD_TYPE.JSON : "JSON",
              FIELD_TYPE.NEWDECIMAL : "NEWDECIMAL",
              FIELD_TYPE.ENUM : "ENUM",
              FIELD_TYPE.SET : "SET", 
              FIELD_TYPE.TINY_BLOB : "TINY_BLOB",
              FIELD_TYPE.MEDIUM_BLOB : "MEDIUM_BLOB",
              FIELD_TYPE.LONG_BLOB : "LONG_BLOB",
              FIELD_TYPE.BLOB : "BLOB",
              FIELD_TYPE.VAR_STRING : "VAR_STRING",
              FIELD_TYPE.STRING : "STRING",
              FIELD_TYPE.GEOMETRY : "GEOMETRY" }

field_flags = {FIELD_FLAG.NOT_NULL : "NOT_NULL",
               FIELD_FLAG.PRIMARY_KEY : "PRIMARY_KEY",
               FIELD_FLAG.UNIQUE_KEY : "UNIQUE_KEY",
               FIELD_FLAG.MULTIPLE_KEY : "PART_KEY",
               FIELD_FLAG.BLOB : "BLOB",
               FIELD_FLAG.UNSIGNED : "UNSIGNED",
               FIELD_FLAG.ZEROFILL : "ZEROFILL",
               FIELD_FLAG.BINARY : "BINARY",
               FIELD_FLAG.ENUM : "NUMERIC",
               FIELD_FLAG.AUTO_INCREMENT : "AUTO_INCREMENT",
               FIELD_FLAG.TIMESTAMP : "TIMESTAMP",
               FIELD_FLAG.SET : "SET",
               FIELD_FLAG.NO_DEFAULT : "NO_DEFAULT",
               FIELD_FLAG.ON_UPDATE_NOW : "UPDATE_TIMESTAMP",
               FIELD_FLAG.NUMERIC : "NUMERIC"}

class fieldinfo():

    def type(self, description):
        if description[1] in field_types:
            return field_types[description[1]]
        return None 

    def flag(self, description):
        flags = [field_flags[f] for f in field_flags.keys() if description[7] & f]
        return " | ".join(flags)



#
# Copyright (C) 2020-2021 Georg Richter and MariaDB Corporation AB

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.

# You should have received a copy of the GNU Library General Public
# License along with this library; if not see <http://www.gnu.org/licenses>
# or write to the Free Software Foundation, Inc.,
# 51 Franklin St., Fifth Floor, Boston, MA 02110, USA
#

import mariadb, collections, datetime
from numbers import Number
from mariadb.constants import *
from typing import Sequence

PARAMSTYLE_QMARK= 1
PARAMSTYLE_FORMAT= 2
PARAMSTYLE_PYFORMAT= 3

RESULT_TUPLE= 0
RESULT_NAMEDTUPLE= 1
RESULT_DICTIONARY= 2

# Command types
SQL_NONE= 0,
SQL_INSERT= 1
SQL_UPDATE= 2
SQL_REPLACE= 3
SQL_DELETE= 4
SQL_CALL= 5
SQL_DO= 6
SQL_SELECT=7
SQL_OTHER=255


class Cursor(mariadb._mariadb.cursor):
    """
    MariaDB Connector/Python Cursor Object
    """

    def check_closed(self):
        if self.closed:
            self._connection._check_closed()
            raise mariadb.ProgrammingError("Cursor is closed")

    def __init__(self, connection, **kwargs):
        """
        initialization
        """
        self._bulk= False
        self._dictionary= False
        self._named_tuple= False
        self._connection= connection
        self._resulttype= RESULT_TUPLE
        self._description= None
        self._transformed_statement= None
        self._prepared= False
        self._prev_stmt= None
        self._force_binary= None
        self._rowcount= 0
        self.buffered= True
        self._parseinfo= None
        self._data= None

        if not connection:
            raise mariadb.ProgrammingError("Invalid or no connection provided")

        # parse keywords
        if kwargs:
             rtype= kwargs.pop("named_tuple", False)
             if rtype:
                 self._resulttype= RESULT_NAMEDTUPLE
             else:
                 rtype= kwargs.pop("dictionary", False)
                 if rtype:
                     self._resulttype= RESULT_DICTIONARY
             buffered= kwargs.pop("buffered", True)
             self.buffered= buffered
             self._prepared= kwargs.pop("prepared", False)
             self._force_binary= kwargs.pop("binary", False)
             self._cursor_type= kwargs.pop("cursor_type", 0)

        # call initialization of main class 
        super().__init__(connection, **kwargs)

                
    def _substitute_parameters(self):
        """
        Internal use only.

        When running in text protocol, this method will replace placeholders
        by supplied values.

        For values which aren't numbers, strings or bytes string representation
        will be used.
        """

        new_stmt= self.statement.encode("utf8")
        replace_diff= 0
        if self._paramlist:
            for i in range (0,len(self._paramlist)):
                if self._paramstyle == PARAMSTYLE_PYFORMAT:
                    val= self._data[self._keys[i]]
                else:
                    val= self._data[i]
                if val is None:
                    replace= "NULL";
                else:
                    if isinstance(val, INDICATOR.MrdbIndicator):
                        if val == INDICATOR.NULL:
                           replace= "NULL"
                        if val == INDICATOR.DEFAULT:
                           replace= "DEFAULT"   
                    elif isinstance(val, Number):
                        replace= val.__str__()
                    else:
                        if isinstance(val, (bytes, bytearray)):
                            replace= "\"%s\"" % self.connection.escape_string(val.decode(encoding='latin1'))
                        else:
                            replace= "\"%s\"" % self.connection.escape_string(val.__str__())
                ofs= self._paramlist[i] + replace_diff
                
                new_stmt= new_stmt[:ofs] + replace.__str__().encode("utf8") + new_stmt[ofs+1:]
                replace_diff+= len(replace) - 1
        return new_stmt

    def _check_execute_params(self):
        # check data format
        if self._paramstyle == PARAMSTYLE_QMARK or \
           self._paramstyle == PARAMSTYLE_FORMAT:
            if not isinstance(self._data, (tuple,list)):
                raise mariadb.ProgrammingError("Data argument must be Tuple or List")

        if self._paramstyle == PARAMSTYLE_PYFORMAT:
            if not isinstance(self._data, dict):
                raise mariadb.ProgrammingError("Data argument must be Dictionary")
            for i in range(0, len(self._keys)):
                if self._keys[i] not in self._data:
                    raise mariadb.ProgrammingError("Dictionary doesn't contain key '%s'" % self._keys[i])
        else:
            # check if number of place holders matches the number of 
            # supplied elements in data tuple
            if self._paramlist and  ((not self._data and len(self._paramlist) > 0) or \
               (len(self._data) != len(self._paramlist))):
                raise mariadb.ProgrammingError("Number of parameters in statement (%s)"\
                                        " doesn't match the number of data elements (%s)."\
                                         % (len(self._paramlist), len(self._data)))

    def callproc(self, sp: str, data: Sequence =()):
        """
        Executes a stored procedure sp. The data sequence must contain an entry for
        each parameter the procedure expects.

        Input/Output or Output parameters have to be retrieved by .fetch methods,
        the .sp_outparams attribute indicates if the result set contains output
        parameters.

        Arguments:
            - sp: Name of stored procedure.
            - data: Optional sequence containing data for placeholder substitution.
        """

        self.check_closed()

        # create statement 
        params= ""
        if data and len(data):
            params= ("?," * len(data))[:-1]
        statement= "CALL %s(%s)" % (sp, params)
        self._rowcount= 0
        self.execute(statement, data)

    def _parse_execute(self, statement:str, data=(), is_bulk=False):
        """
        For internal use

        Parses SQL statement and checks parameters.
        """

        if not statement:
            raise mariadb.ProgrammingError("empty statement")

        # parse statement
        if self.statement != statement or is_bulk and not self._bulk:
            super()._parse(statement)
            self._prev_stmt= statement
            self._reprepare= True
        else:
            self._reprepare= False

        self._transformed_statement= self.statement

        if self._cursor_type == CURSOR.READ_ONLY:
            self._text= False

        self._data= data

        self._check_execute_params()
       

    def nextset(self):
        """
        Will make the cursor skip to the next available result set,
        discarding any remaining rows from the current set.
        """

        self.check_closed()
        return super()._nextset()

    def execute(self, statement: str, data: Sequence =(), buffered=False):
        """
        Prepare and execute a SQL statement.

        Parameters may be provided as sequence or mapping and will be bound
        to variables in the operation. Variables are specified as question
        marks (paramstyle='qmark'), however for compatibility reasons MariaDB
        Connector/Python also supports the 'format' and 'pyformat' paramstyles
        with the restriction, that different paramstyles can't be mixed within
        a statement.

        A reference to the operation will be retained by the cursor.
        If the cursor was created with attribute prepared=True the statement
        string for following execute operations will be ignored.
        This is most effective for algorithms where the same operation is used,
        but different parameters are bound to it (many times).

        By default execute() method generates an unbuffered result set for 
        statements which return data, setting optional parameter buffered to
        True will generate buffered result sets.
        """

        self.check_closed()

        # Parse statement
        do_parse= True
        self._rowcount= 0

        if buffered:
            self.buffered= True

        # clear pending result sets
        if self.field_count:
            self._clear_result()

        # if we have a prepared cursor, we have to set statement
        # to previous statement and don't need to parse
        if self._prepared and self.statement:
            statement= self.statement
            do_parse= False

        # parse statement and check param style
        if do_parse:
            self._parse_execute(statement, (data))

        self._description= None

        if len(data):
            self._data= data
        else:
            self._data= None
            # If statement doesn't contain parameters we force to run in text
            # mode, unless a server side cursor or stored procedure will be
            # executed.
            if self._command != SQL_CALL and self._cursor_type == 0:
                self._text= True

        if self._force_binary:
           self._text= False

        for val in data:
            if isinstance(val, (bytes, bytearray, datetime.datetime, datetime.date, datetime.time)):
                self._text= False
                break

        if self._text:
            # in text mode we need to substitute parameters
            # and store transformed statement
            if (self.paramcount > 0):
                 self._transformed_statement= self._substitute_parameters()
            else:
                 self._transformed_statement= self.statement

            self._execute_text(self._transformed_statement)
            self._readresponse()
        else:
            self._data= data
            self._execute_binary()

        self._initresult()
        self._bulk= 0

    def executemany(self, statement, parameters):
        """ 
        Prepare a database operation (INSERT,UPDATE,REPLACE or DELETE statement) and
        execute it against all parameter found in sequence.

        Exactly behaves like .execute() but accepts a list of tuples, where each
        tuple represents data of a row within a table.
        .executemany() only supports DML (insert, update, delete) statements.
        """
        self.check_closed()

        if not parameters or not len(parameters):
            raise mariadb.ProgrammingError("No data provided")

        # clear pending results
        if self.field_count:
            self._clear_result()

        # If the server doesn't support bulk operations, we need to emulate
        # by looping
        # TODO: insert/replace statements are not optimized yet, rowcount not set
        if not (self.connection.extended_server_capabilities & (CLIENT.BULK_OPERATIONS >> 32)):
            count= 0
            for row in parameters:
                self.execute(statement, row)
                count+= self.rowcount
            self._rowcount= count
        else:
            # parse statement 
            self._parse_execute(statement, parameters[0], is_bulk=True)
            self._data= parameters
            self.is_text= False
            self._rowcount= 0
            self._execute_bulk()
            self._bulk= 1

    def _fetch_row(self):
        """
        Internal use only

        fetches row and converts values, if connections has a converter.
        """
        self.check_closed()

        # if there is no result set, PEP-249 requires to raise an
        # exception
        if not self.field_count:
             raise mariadb.ProgrammingError("Cursor doesn't have a result set")
        row= super().fetchone()
        if self._connection._converter:
            l= list(row)
            if not self._description:
                self._description= super().description
            for i,v in enumerate(row):
                type= self.description[i][1]
                if type in self._connection._converter:
                    func= self._connection._converter[type]
                    l[i]= func(v)
                else:
                    l[i]= v
            row= tuple(l)
        return row

    def close(self):
        """
        Closes the cursor. 

        If the cursor has pending or unread results, .close() will cancel them
        so that further operations using the same connection can be executed.

        The cursor will be unusable from this point forward; an Error (or subclass)
        exception will be raised if any operation is attempted with the cursor."
        """
        super().close()

    def fetchone(self):
        """
        Fetch the next row of a query result set, returning a single sequence,
        or None if no more data is available.

        An exception will be raised if the previous call to execute() didn't 
        produce a result set or execute() wasn't called before.
        """
        self.check_closed()

        row= self._fetch_row()
        if not row:
            return row
        if self._resulttype == RESULT_DICTIONARY:
            ret= dict(zip(list(d[0] for d in self.description),row))
        elif self._resulttype == RESULT_NAMEDTUPLE:
            ret= collections.namedtuple('Row1', list(d[0] for d in self.description));
            ret= ret._make(row)
        else:
            ret= row
        return ret

    def fetchmany(self, size: int =0):
        """
        Fetch the next set of rows of a query result, returning a sequence
        of sequences (e.g. a list of tuples). An empty sequence is returned
        when no more rows are available.

        The number of rows to fetch per call is specified by the parameter. 
        If it is not given, the cursor's arraysize determines the number
        of rows to be fetched. The method should try to fetch as many rows 
        as indicated by the size parameter. 
        If this is not possible due to the specified number of rows not being
        available, fewer rows may be returned.

        An exception will be raised if the previous call to execute() didn't 
        produce a result set or execute() wasn't called before.
        """
        self.check_closed()

        rows=[]
        if size == 0:
            size= self.arraysize
        for count in range(0, size):
            row= self.fetchone()
            if row:
                rows.append(row)
        return rows

    def fetchall(self):
        """
        Fetch all remaining rows of a query result, returning them as a 
        sequence of sequences (e.g. a list of tuples).

        An exception will be raised if the previous call to execute() didn't 
        produce a result set or execute() wasn't called before.
        """
        self.check_closed()

        rows=[];
        for row in self:
            rows.append((row))
        return rows

    def scroll(self, value: int, mode="relative"):
        """
        Scroll the cursor in the result set to a new position according to mode.

        If mode is "relative" (default), value is taken as offset to the current
        position in the result set, if set to absolute, value states an absolute
        target position.
        """

        if self.field_count == 0:
            raise mariadb.ProgrammingError("Cursor doesn't have a result set")

        if not self.buffered:
            raise mariadb.ProgrammingError("This method is available only for cursors "\
                                           "with a buffered result set.")

        if mode != "absolute" and mode != "relative":
            raise mariadb.ProgrammingError("Invalid or unknown scroll mode specified.")

        if value == 0 and mode != "absolute":
            raise mariadb.ProgrammingError("Invalid position value 0.")

        if mode == "relative":
            if self.rownumber + value < 0 or \
               self.rownumber + value > self.rowcount:
                raise mariadb.ProgrammingError("Position value is out of range.")
            new_pos= self.rownumber + value
        else:
            if value < 0 or value >= self.rowcount:
                raise mariadb.ProgrammingError("Position value is out of range.")
            new_pos= value

        self._seek(new_pos);
        self._rownumber= new_pos;

    def setinputsizes(self, size: int):
        """
        Required by PEP-249. Does nothing in MariaDB Connector/Python
        """

        return

    def setoutputsize(self, size: int):
        """
        Required by PEP-249. Does nothing in MariaDB Connector/Python
        """

        return

    def __enter__(self):
        """Returns a copy of the cursor."""

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Closes cursor."""
        self.close()

    def __del__(self):
        self.close()

    @property
    def rowcount(self):
        """
        This read-only attribute specifies the number of rows that the last\
        execute*() produced (for DQL statements like SELECT) or affected
        (for DML statements like UPDATE or INSERT).
        The return value is -1 in case no .execute*() has been performed
        on the cursor or the rowcount of the last operation  cannot be
        determined by the interface.
        """
        self.check_closed()
        if self._rowcount > 0:
            return self._rowcount
        return super().rowcount

    @property
    def sp_outparams(self):
        """
        Indicates if the current result set contains in out or out parameter
        from a previous executed stored procedure
        """
        self.check_closed()

        return bool(self.connection.server_status & STATUS.PS_OUT_PARAMS)

    @property
    def lastrowid(self):
        """
        Returns the ID generated by a query on a table with a column having
        the AUTO_INCREMENT attribute or the value for the last usage of
        LAST_INSERT_ID().

        If the last query wasn't an INSERT or UPDATE
        statement or if the modified table does not have a column with the
        AUTO_INCREMENT attribute and LAST_INSERT_ID was not used, the returned
        value will be zero
        """
        self.check_closed()

        id= self.insert_id
        if id > 0:
            return id
        return None

    @property
    def connection(self):
        """
        Read-Only attribute which returns the reference to the connection
        object on which the cursor was created.
        """
        self.check_closed()

        return self._connection



"""
    Cursor constants are used for server side cursors.
    Currently only read only cursor is supported.

    Cursor constants are defined in module *mariadb.constants.CURSOR*.
"""

NONE= 0
READ_ONLY= 1



"""
Constants for _get_info method of MariadB connection object
"""

CHARSET_ID= 0
CHARSET_NAME= 1 
CLIENT_ERRORS= 2
CLIENT_VERSION= 3
CLIENT_VERSION_ID= 4
ASYNC_TIMEOUT= 5
ASYNC_TIMEOUT_MS= 6
CHARSET_INFO= 7
ERROR= 8
ERROR_ID= 9
HOST= 10
INFO= 11
PORT= 12
PROTOCOL_VERSION_ID= 13
PVIO_TYPE= 14
SCHEMA= 15
SERVER_TYPE= 16
SERVER_VERSION= 17 
SERVER_VERSION_ID= 18 
SOCKET= 19
SQLSTATE= 20
SSL_CIPHER= 21
TLS_LIBRARY= 22
TLS_VERSION= 23
TLS_VERSION_ID= 24
TYPE= 25
UNIX_SOCKET= 26
USER= 27
MAX_ALLOWED_PACKET= 28
NET_BUFFER_LENGTH= 29
SERVER_STATUS= 30
SERVER_CAPABILITIES= 31
EXTENDED_SERVER_CAPABILITIES= 32
CLIENT_CAPABILITIES= 33
BYTES_READ= 34
BYTES_SENT= 35



#
# Copyright (C) 2020-2021 Georg Richter and MariaDB Corporation AB

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.

# You should have received a copy of the GNU Library General Public
# License along with this library; if not see <http://www.gnu.org/licenses>
# or write to the Free Software Foundation, Inc.,
# 51 Franklin St., Fifth Floor, Boston, MA 02110, USA
#

import mariadb
import socket
import time
import mariadb.cursors

from mariadb.constants import STATUS, TPC_STATE, INFO
from distutils.version import StrictVersion

_DEFAULT_CHARSET = "utf8mb4"
_DEFAULT_COLLATION = "utf8mb4_general_ci"
_MAX_TPC_XID_SIZE=64

class Connection(mariadb._mariadb.connection):
    """
    MariaDB Connector/Python Connection Object

    Handles the connection to a MariaDB or MySQL database server.
    It encapsulates a database session.

    Connections are created using the method mariadb.connect()
    """

    def _check_closed(self):
        if self._closed:
            raise mariadb.ProgrammingError("Invalid connection or not connected")

    def __init__(self, *args, **kwargs):
        """
        Establishes a connection to a database server and returns a connection
        object.
        """

        self._socket= None
        self.__in_use= 0
        self.__pool = None
        self.__last_used = 0
        self.tpc_state= TPC_STATE.NONE
        self._xid= None

        autocommit= kwargs.pop("autocommit", False)
        self._converter= kwargs.pop("converter", None)

        # if host contains a connection string or multiple hosts,
        # we need to check if it's supported by Connector/C
        if "host" in kwargs:
            host= kwargs.get("host")
            if StrictVersion(mariadb.mariadbapi_version) < StrictVersion('3.3.0') and ',' in host:
                raise mariadb.ProgrammingError("Host failover list requires MariaDB Connector/C 3.3.0 or newer")


        # compatibility feature: if SSL is provided as a dictionary,
        # we will map it's content
        if "ssl" in kwargs and not isinstance(kwargs["ssl"], bool):
             ssl= kwargs.pop("ssl", None)
             for key in ["ca", "cert", "capath", "key", "cipher"]:
                 if key in ssl:
                     kwargs["ssl_%s" % key] = ssl[key]
             kwargs["ssl"]= True

        super().__init__(*args, **kwargs)
        self.autocommit= autocommit

    def cursor(self, cursorclass=mariadb.cursors.Cursor, **kwargs):
        """
        Returns a new cursor object for the current connection.

        If no cursorclass was specified, a cursor with default mariadb.Cursor class
        will be created.

        Optional keyword parameters:

        - buffered= True
          If set to False the result will be unbuffered, which means before executing
          another statement with the same connection the entire result set must be fetched.
          Please note that the default was False for MariaDB Connector/Python versions < 1.1.0.
        - dictionary= False
          Return fetch values as dictionary.

        - named_tuple= False
          Return fetch values as named tuple. This feature exists for compatibility reasons
          and should be avoided due to possible inconsistency.

        - cursor_type= CURSOR_TYPE.NONE
          If cursor_type is set to CURSOR_TYPE.READ_ONLY, a cursor is opened for the
          statement invoked with cursors execute() method.

        - prepared= False
          When set to True cursor will remain in prepared state after the first execute()
          method was called. Further calls to execute() method will ignore the sql statement.

        - binary= False
          Always execute statement in MariaDB client/server binary protocol.

        In versions prior to 1.1.0 results were unbuffered by default, which means before executing another 
        statement with the same connection the entire result set must be fetched.

        fetch* methods of the cursor class by default return result set values as a 
        tuple, unless named_tuple or dictionary was specified. The latter one exists 
        for compatibility reasons and should be avoided due to possible inconsistency
        in case two or more fields in a result set have the same name.

        If cursor_type is set to CURSOR_TYPE.READ_ONLY, a cursor is opened for
        the statement invoked with cursors execute() method.
        """
        self._check_closed()
        cursor= cursorclass(self, **kwargs)
        if not isinstance(cursor, mariadb._mariadb.cursor):
            raise mariadb.ProgrammingError("%s is not an instance of mariadb.cursor" % cursor)
        return cursor

    def close(self):
        self._check_closed()
        if self._Connection__pool:
            self._Connection__pool._close_connection(self)
        else:
            super().close()

    def __enter__(self):
        self._check_closed()
        "Returns a copy of the connection."

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._check_closed()
        "Closes connection."

        self.close()

    def commit(self):
        """
        Commit any pending transaction to the database.
        """

        self._check_closed()
        if self.tpc_state > TPC_STATE.NONE:
            raise mariadb.ProgrammingError("commit() is not allowed if a TPC transaction is active")
        self._execute_command("COMMIT")
        self._read_response()

    def rollback(self):
        """
        Causes the database to roll back to the start of any pending transaction

        Closing a connection without committing the changes first will cause an
        implicit rollback to be performed.
        Note that rollback() will not work as expected if autocommit mode was set to True
        or the storage engine does not support transactions."
        """

        self._check_closed()
        if self.tpc_state > TPC_STATE.NONE:
            raise mariadb.ProgrammingError("rollback() is not allowed if a TPC transaction is active")
        self._execute_command("ROLLBACK")
        self._read_response()

    def kill(self, id: int):
        """
        This function is used to ask the server to kill a database connection
        specified by the processid parameter. 

        The connection id can be be retrieved by SHOW PROCESSLIST sql command.
        """

        self._check_closed()
        if not isinstance(id, int):
            raise mariadb.ProgrammingError("id must be of type int.")
        stmt= "KILL %s" % id
        self._execute_command(stmt)
        self._read_response()

    def begin(self):
        """
        Start a new transaction which can be committed by .commit() method,
        or cancelled by .rollback() method.
        """
        self._check_closed()
        self._execute_command("BEGIN")
        self._read_response()

    def select_db(self, new_db: str):
        """
        Gets the default database for the current connection.

        The default database can also be obtained or changed by database attribute.
        """

        self._check_closed()
        self.database= new_db

    def get_server_version(self):
        """
        Returns a tuple representing the version of the connected server in
        the following format: (MAJOR_VERSION, MINOR_VERSION, PATCH_VERSION)
        """

        return self.server_version_info

    def show_warnings(self):
        """
        Shows error, warning and note messages from last executed command.
        """

        self._check_closed()
        if (not self.warnings):
            return None;

        cursor= self.cursor()
        cursor.execute("SHOW WARNINGS")
        ret= cursor.fetchall()
        del cursor
        return ret

    class xid(tuple):
        """
        xid(format_id: int, global_transaction_id: str, branch_qualifier: str)

        Creates a transaction ID object suitable for passing to the .tpc_*()
        methods of this connection.

        Parameters:

        - format_id: Format id. If not set default value `0` will be used.

        - global_transaction_id: Global transaction qualifier, which must be unique. The maximum length of the global transaction id is limited to 64 characters.

        - branch_qualifier: Branch qualifier which represents a local transaction identifier. The maximum length of the branch qualifier is limited to 64 characters.
 
        """
        def __new__(self, format_id, transaction_id, branch_qualifier):
            if not isinstance(format_id, int):
                raise mariadb.ProgrammingError("argument 1 must be int, not %s", type(format_id).__name__)
            if not isinstance(transaction_id, str):
                raise mariadb.ProgrammingError("argument 2 must be str, not %s", type(transaction_id).__mane__)
            if not isinstance(branch_qualifier, str):
                raise mariadb.ProgrammingError("argument 3 must be str, not %s", type(transaction_id).__name__)
            if len(transaction_id) > _MAX_TPC_XID_SIZE:
                raise mariadb.ProgrammingError("Maximum length of transaction_id exceeded.")
            if len(branch_qualifier) > _MAX_TPC_XID_SIZE:
                raise mariadb.ProgrammingError("Maximum length of branch_qualifier exceeded.")
            if format_id == 0:
                format_id= 1
            return super().__new__(self, (format_id, transaction_id, branch_qualifier))

    def tpc_begin(self, xid):
        """
        Parameter:
          xid: xid object which was created by .xid() method of connection class

        Begins a TPC transaction with the given transaction ID xid.

        This method should be called outside of a transaction
        (i.e. nothing may have executed since the last .commit()
        or .rollback()).
        Furthermore, it is an error to call .commit() or .rollback() within
        the TPC transaction. A ProgrammingError is raised, if the application
        calls .commit() or .rollback() during an active TPC transaction.
        """

        self._check_closed()
        if type(xid).__name__ != "xid":
            raise mariadb.ProgrammingError("argument 1 must be xid not %s", type(xid).__name__)
        stmt= "XA BEGIN '%s','%s',%s" % (xid[1], xid[2], xid[0])
        try:
            self._execute_command(stmt)
            self._read_response()
        except:
            raise
        self.tpc_state= TPC_STATE.XID
        self._xid= xid

    def tpc_commit(self, xid=None):
        """
        Optional parameter:"
        xid: xid object which was created by .xid() method of connection class.

        When called with no arguments, .tpc_commit() commits a TPC transaction
        previously prepared with .tpc_prepare().

        If .tpc_commit() is called prior to .tpc_prepare(), a single phase commit
        is performed. A transaction manager may choose to do this if only a
        single resource is participating in the global transaction.
        When called with a transaction ID xid, the database commits the given
        transaction. If an invalid transaction ID is provided, a ProgrammingError
        will be raised. This form should be called outside of a transaction, and
        is intended for use in recovery."
        """

        self._check_closed()
        if not xid:
            xid= self._xid

        if self.tpc_state == TPC_STATE.NONE:
            raise mariadb.ProgrammingError("Transaction not started.")
        if xid is None and self.tpc_state != TPC_STATE.PREPARE:
            raise mariadb.ProgrammingError("Transaction is not prepared.")
        if xid and type(xid).__name__ != "xid":
            raise mariadb.ProgrammingError("argument 1 must be xid not %s" % type(xid).__name__)

        if self.tpc_state < TPC_STATE.PREPARE:
            stmt= "XA END '%s','%s',%s" % (xid[1], xid[2], xid[0])
            self._execute_command(stmt)
            try:
                self._read_response()
            except:
                self._xid= None
                self.tpc_state= TPC_STATE.NONE
                raise

        stmt= "XA COMMIT '%s','%s',%s" % (xid[1], xid[2], xid[0])
        if self.tpc_state < TPC_STATE.PREPARE:
            stmt= stmt + " ONE PHASE"
        try:
            self._execute_command(stmt)
            self._read_response()
        except:
            self._xid= None
            self.tpc_state= TPC_STATE.NONE
            raise

        #cleanup
        self._xid= None
        self.tpc_state= TPC_STATE.NONE

    def tpc_prepare(self): 
        """
        Performs the first phase of a transaction started with .tpc_begin().
        A ProgrammingError will be raised if this method was called outside of
        a TPC transaction.

        After calling .tpc_prepare(), no statements can be executed until
        .tpc_commit() or .tpc_rollback() have been called.
        """

        self._check_closed()
        if self.tpc_state == TPC_STATE.NONE:
            raise mariadb.ProgrammingError("Transaction not started.")
        if self.tpc_state == TPC_STATE.PREPARE:
            raise mariadb.ProgrammingError("Transaction is already in prepared state.")

        xid= self._xid
        stmt= "XA END '%s','%s',%s" % (xid[1], xid[2], xid[0])
        try:
            self._execute_command(stmt)
            self._read_response()
        except:
            self._xid= None
            self.tpc_state= TPC_STATE.NONE
            raise

        stmt= "XA PREPARE '%s','%s',%s" % (xid[1], xid[2], xid[0])
        try:
            self._execute_command(stmt)
            self._read_response()
        except:
            self._xid= None
            self.tpc_state= TPC_STATE.NONE
            raise

        self.tpc_state= TPC_STATE.PREPARE

    def tpc_rollback(self, xid=None):
        """
        Parameter:
           xid: xid object which was created by .xid() method of connection
                class

        Performs the first phase of a transaction started with .tpc_begin().
        A ProgrammingError will be raised if this method outside of a TPC
        transaction.

        After calling .tpc_prepare(), no statements can be executed until
        .tpc_commit() or .tpc_rollback() have been called.
        """

        self._check_closed()
        if self.tpc_state == TPC_STATE.NONE:
            raise mariadb.ProgrammingError("Transaction not started.")
        if xid and type(xid).__name__ != "xid":
            raise mariadb.ProgrammingError("argument 1 must be xid not %s" % type(xid).__name__)

        if not xid:
            xid= self._xid

        if self.tpc_state < TPC_STATE.PREPARE:
            stmt= "XA END '%s','%s',%s" % (xid[1], xid[2], xid[0])
            self._execute_command(stmt)
            try:
                self._read_response()
            except:
                self._xid= None
                self.tpc_state= TPC_STATE.NONE
                raise

        stmt= "XA ROLLBACK '%s','%s',%s" % (xid[1], xid[2], xid[0])
        try:
            self._execute_command(stmt)
            self._read_response()
        except:
            self._xid= None
            self.tpc_state= TPC_STATE.NONE
            raise

        self.tpc_state= TPC_STATE.PREPARE

    def tpc_recover(self):
        """
        Returns a list of pending transaction IDs suitable for use with
        tpc_commit(xid) or .tpc_rollback(xid).
        """

        self._check_closed()
        cursor= self.cursor()
        cursor.execute("XA RECOVER")
        result= cursor.fetchall()
        del cursor
        return result

    @property
    def database(self):
        """Get default database for connection."""

        self._check_closed()
        return self._mariadb_get_info(INFO.SCHEMA, str)
 
    @database.setter
    def database(self, schema):
        """Set default database."""
        self._check_closed()

        try:
            self._execute_command("USE %s" % str(schema))
            self._read_response()
        except:
            raise

    @property
    def user(self):
        """
        Returns the user name for the current connection or empty
        string if it can't be determined, e.g. when using socket
        authentication.
        """
        self._check_closed()

        return self._mariadb_get_info(INFO.USER, str)

    @property
    def character_set(self):
        """
        Client character set.

        For MariaDB Connector/Python it is always utf8mb4.
        """

        return _DEFAULT_CHARSET

    @property
    def client_capabilities(self):
        """Client capability flags."""

        self._check_closed()
        return self._mariadb_get_info(INFO.CLIENT_CAPABILITIES, int)

    @property
    def server_capabilities(self):
        """Server capability flags."""

        self._check_closed()
        return self._mariadb_get_info(INFO.SERVER_CAPABILITIES, int)

    @property
    def extended_server_capabilities(self):
        """Extended server capability flags (only for MariaDB database servers)."""

        self._check_closed()
        return self._mariadb_get_info(INFO.EXTENDED_SERVER_CAPABILITIES, int)

    @property
    def server_port(self):
        """Database server TCP/IP port. This value will be 0 in case of a unix socket connection."""

        self._check_closed()
        return self._mariadb_get_info(INFO.PORT, int)

    @property
    def unix_socket(self):
        """Unix socket name."""

        self._check_closed()
        return self._mariadb_get_info(INFO.UNIX_SOCKET, str)

    @property
    def server_name(self):
        """Name or IP address of database server."""

        self._check_closed()
        return self._mariadb_get_info(INFO.HOST, str)

    @property
    def collation(self):
        """Client character set collation"""

        return _DEFAULT_COLLATION

    @property
    def server_info(self):
        """Server version in alphanumerical format (str)"""

        self._check_closed()
        return self._mariadb_get_info(INFO.SERVER_VERSION, str)

    @property
    def tls_cipher(self):
        """TLS cipher suite if a secure connection is used."""

        self._check_closed()
        return self._mariadb_get_info(INFO.SSL_CIPHER, str)

    @property
    def tls_version(self):
        """TLS protocol version if a secure connection is used."""

        self._check_closed()
        return self._mariadb_get_info(INFO.TLS_VERSION, str)

    @property
    def server_status(self):
        """
        Return server status flags
        """

        self._check_closed()
        return self._mariadb_get_info(INFO.SERVER_STATUS, int)

    @property
    def server_version(self):
        """
        Server version in numerical format.

        The form of the version number is
        VERSION_MAJOR * 10000 + VERSION_MINOR * 100 + VERSION_PATCH
        """

        self._check_closed()
        return self._mariadb_get_info(INFO.SERVER_VERSION_ID, int)

    @property
    def server_version_info(self):
        """
        Returns numeric version of connected database server in tuple format. 
        """

        self._check_closed()
        version= self.server_version
        return (int(version / 10000), int((version % 10000) / 100), version % 100)

    @property
    def autocommit(self):
        """
        Toggles autocommit mode on or off for the current database connection.

        Autocommit mode only affects operations on transactional table types.
        Be aware that rollback() will not work, if autocommit mode was switched
        on.

        By default autocommit mode is set to False."
        """

        self._check_closed()
        return bool(self.server_status & STATUS.AUTOCOMMIT)

    @autocommit.setter
    def autocommit(self, mode):
        self._check_closed()
        if bool(mode) == self.autocommit:
            return
        try:
            self._execute_command("SET AUTOCOMMIT=%s" % int(mode))
            self._read_response()
        except:
            raise

    @property
    def socket(self):
        """Returns the socket used for database connection"""

        fno= self.get_socket()
        if not self._socket:
            self._socket= socket.socket(fileno=fno)
        # in case of a possible reconnect, file descriptor has changed
        elif fno != self._socket.fileno():
            self._socket= socket.socket(fileno=fno)
        return self._socket

    @property
    def open(self):
        """
        Returns true if the connection is alive.

        A ping command will be send to the server for this purpose,
        which means this function might fail if there are still
        non processed pending result sets.
        """

        self._check_closed()
        try:
            self.ping()
        except:
            return False
        return True

    # Aliases
    character_set_name= character_set

    @property
    def thread_id(self):
        """
        Alias for connection_id
        """

        self._check_closed()
        return self.connection_id



"""MariaDB FIELD_FLAG Constants

These constants represent the various field flags. As an addition
to the DBAPI 2.0 standard (PEP-249) these flags are returned as
eighth element of the cursor description attribute.

Field flags are defined in module *mariadb.constants.FIELD_FLAG*
"""

# Source: mariadb_com.h (MariaDB Connector(C)

NOT_NULL = 1
PRIMARY_KEY = 2
UNIQUE_KEY = 4
MULTIPLE_KEY = 8
BLOB = 16
UNSIGNED = 32
ZEROFILL = 64
BINARY = 128
ENUM = 256
AUTO_INCREMENT = 512
TIMESTAMP = 1024
SET = 2048
NO_DEFAULT = 4096
ON_UPDATE_NOW = 8192
NUMERIC = 32768
PART_OF_KEY = 16384
GROUP = 32768
UNIQUE = 65536	



'''
MariaDB indicator variables

Indicator values are used in executemany() method of cursor class to
indicate special values.
'''
class MrdbIndicator():
    indicator= 0

    def __init__(self, indicator):
        self.indicator = indicator

NULL = MrdbIndicator(1)
DEFAULT = MrdbIndicator(2)
IGNORE = MrdbIndicator(3)
IGNORE_ROW = MrdbIndicator(4)



#!/usr/bin/env python

import os

from setuptools import setup, Extension
from configparser import ConfigParser

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

define_macros= []

# read settings from site.cfg
c= ConfigParser()
c.read(['site.cfg'])
options= dict(c.items('cc_options'))

if os.name == "posix":
    from mariadb_posix import get_config
if os.name == "nt":
    from mariadb_windows import get_config

cfg = get_config(options)

PY_MARIADB_AUTHORS= "Georg Richter"

PY_MARIADB_MAJOR_VERSION=1
PY_MARIADB_MINOR_VERSION=1
PY_MARIADB_PATCH_VERSION=2
PY_MARIADB_PRE_RELEASE_SEGMENT=""

PY_MARIADB_VERSION= "%s.%s.%s" % (PY_MARIADB_MAJOR_VERSION, PY_MARIADB_MINOR_VERSION, PY_MARIADB_PATCH_VERSION)

PY_MARIADB_PRE_RELEASE_NR=0
PY_MARIADB_VERSION_INFO= (PY_MARIADB_MAJOR_VERSION, PY_MARIADB_MINOR_VERSION, PY_MARIADB_PATCH_VERSION,
                          PY_MARIADB_PRE_RELEASE_SEGMENT, PY_MARIADB_PRE_RELEASE_NR)

define_macros.append(("PY_MARIADB_MAJOR_VERSION", PY_MARIADB_MAJOR_VERSION))
define_macros.append(("PY_MARIADB_MINOR_VERSION", PY_MARIADB_MINOR_VERSION))
define_macros.append(("PY_MARIADB_PATCH_VERSION", PY_MARIADB_PATCH_VERSION))
define_macros.append(("PY_MARIADB_PRE_RELEASE_SEGMENT", "\"%s\"" % PY_MARIADB_PRE_RELEASE_SEGMENT))


with open("mariadb/release_info.py", "w") as rel_info:
   rel_info.write("__author__='%s'\n__version__='%s'\n__version_info__=%s" %
             (PY_MARIADB_AUTHORS, PY_MARIADB_VERSION, PY_MARIADB_VERSION_INFO))

setup(name='mariadb',
      version=PY_MARIADB_VERSION,
      python_requires='>=3.7',
      classifiers = [
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Environment :: MacOS X',
          'Environment :: Win32 (MS Windows)',
          'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
          'Programming Language :: C',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: MacOS',
          'Operating System :: POSIX',
          'Intended Audience :: End Users/Desktop',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'Topic :: Database'
      ],
      description='Python MariaDB extension',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author=PY_MARIADB_AUTHORS,
      license='LGPL 2.1',
      url='https://www.github.com/mariadb-corporation/mariadb-connector-python',
      project_urls={
         "Bug Tracker": "https://jira.mariadb.org/",
         "Documentation": "https://mariadb-corporation.github.io/mariadb-connector-python/",
         "Source Code": "https://www.github.com/mariadb-corporation/mariadb-connector-python",
      },
      ext_modules=[Extension('mariadb._mariadb', ['mariadb/mariadb.c', 'mariadb/mariadb_connection.c',
                                         'mariadb/mariadb_exception.c', 'mariadb/mariadb_cursor.c',
                                         'mariadb/mariadb_codecs.c',
                                         'mariadb/mariadb_parser.c'],
                             define_macros= define_macros,
                             include_dirs=cfg.includes,
                             library_dirs=cfg.lib_dirs,
                             libraries=cfg.libs,
                             extra_compile_args = cfg.extra_compile_args,
                             extra_link_args = cfg.extra_link_args,
                             extra_objects= cfg.extra_objects
                             )],
      py_modules=['mariadb.__init__', 'mariadb.constants.CLIENT', 'mariadb.constants.CURSOR',
                  'mariadb.constants.STATUS', 'mariadb.constants.TPC_STATE', 'mariadb.constants.INFO',
                  'mariadb.field', 'mariadb.connections', 'mariadb.connectionpool', 'mariadb.dbapi20',
                  'mariadb.cursors', 'mariadb.release_info', 'mariadb.constants.ERR',
                  'mariadb.constants.FIELD_TYPE', 'mariadb.constants.FIELD_FLAG', 'mariadb.constants.INDICATOR'],
      )



''' 
MariaDB client flags (internal use only).

These flags are used when establishing a connection to a MariaDB
database server or to check the capabilities of a MariaDB server.

Client flags are defined in module *mariadb.constants.CLIENT_FLAG*
'''

MYSQL = 1  # MariaDB
LONG_PASSWORD= 1 # MySQL
FOUND_ROWS = 2
LONG_FLAG = 4
CONNECT_WITH_DB	= 8
NO_SCHEMA = 16
COMPRESS = 32
LOCAL_FILES = 128
IGNORE_SPACE = 256
INTERACTIVE	= 1024
SSL = 2048
TRANSACTIONS = 8192
SECURE_CONNECTION = 32768  
MULTI_STATEMENTS = 1 << 16
MULTI_RESULTS = 1 << 17
PS_MULTI_RESULTS = 1 << 18
PLUGIN_AUTH = 1 << 19
CONNECT_ATTRS = 1 << 20
CAN_HANDLE_EXPIRED_PASSWORDS = 1 < 22
SESSION_TRACKING = 1 << 23
SSL_VERIFY_SERVER_CERT = 1 << 30
REMEMBER_OPTIONS = 1 << 31

# MariaDB specific capabilities
PROGRESS = 1 << 32
BULK_OPERATIONS = 1 << 34
EXTENDED_METADATA = 1 << 35
CACHE_METDATA = 1 << 36



'''
MariaDB Connector/Python module enables python programs to access MariaDB and MySQL
databases, using an API which is compliant with the Python DB API 2.0 (PEP-249).
'''

from ._mariadb import (
    DataError,
    DatabaseError,
    Error,
    IntegrityError,
    InterfaceError,
    InternalError,
    NotSupportedError,
    OperationalError,
    PoolError,
    ProgrammingError,
    Warning,
    mariadbapi_version,
)

from .field import fieldinfo

_POOLS= _CONNECTION_POOLS= {}

from mariadb.dbapi20 import *
from mariadb.connectionpool import *
from mariadb.cursors import Cursor
from mariadb.release_info import __version__ as __version__ 
from mariadb.release_info import __version_info__ as __version_info__ 
from mariadb.release_info import __author__ as __author__
from mariadb.connections import Connection
# disable for now, until tests are in place
# from mariadb.pooling import *

def connect(*args, connectionclass= mariadb.connections.Connection, **kwargs):
    """
    Creates a MariaDB Connection object.

    By default the standard connectionclass mariadb.connections.Connection will
    be created.

    Parameter connectionclass specifies a subclass of mariadb.Connection object. If not
    specified default will be used. This optional parameter was added in version 1.1.0.

    Connection parameters are provided as a set of keyword arguments:
        - host:
            The host name or IP address of the database server.
            If MariaDB Connector/Python was built with MariaDB Connector/C 3.3 it is also
            possible to provide a comma separated list of hosts for simple fail over in
            case of one or more hosts are not available.
        - user, username:
            The username used to authenticate with the database server
        - password, passwd:
            The password of the given user
        - database, db:
            database (schema) name to use when connecting with the database server
        - unix_socket:
            The location of the unix socket file to use instead of using an IP port
            to connect. If socket authentication is enabled, this can also be used
            in place of a password.
        - port:
            port number of the database server. If not specified the default
            value of 3306 will be used.
        - connect_timeout:
            connect timeout in seconds
        - read_timeout:
            read timeout in seconds
        - write_timeout:
            write timeout in seconds
        - local_infile:
            Enables or disables the use of LOAD DATA LOCAL INFILE statements.
        - compress= False:
            Uses the compressed protocol for client server communication. If the
            server doesn't support compressed protocol, the default protocol will
            be used.
        - init_command:
            Command(s) which will be executed when connecting and reconnecting to
            the database server
        - default_file:
            Read options from the specified option file. If the file is an empty
            string, default configuration file(s) will be used
        - default_group:
            Read options from the specified group
        - plugin_dir:
            Directory which contains MariaDB client plugins.
        - ssl_key:
            Defines a path to a private key file to use for TLS. This option
            requires that you use the absolute path, not a relative path. The
            specified key must be in PEM format
        - ssl_cert:
            Defines a path to the X509 certificate file to use for TLS.
            This option requires that you use the absolute path, not a relative
            path. The X609 certificate must be in PEM format.
        - ssl_ca:
            Defines a path to a PEM file that should contain one or more X509
            certificates for trusted Certificate Authorities (CAs) to use for TLS.
            This option requires that you use the absolute path, not a relative
            path.
        - ssl_capath:
            Defines a path to a directory that contains one or more PEM files that
            contains one X509 certificate for a trusted Certificate Authority (CA)
        - ssl_cipher:
            Defines a list of permitted cipher suites to use for TLS
        - ssl_crlpath:
            Defines a path to a PEM file that should contain one or more revoked
            X509 certificates to use for TLS. This option requires that you use
            the absolute path, not a relative path.
        - ssl_verify_cert:
            Enables server certificate verification.
        - ssl:
            The connection must use TLS security or it will fail.
        - autocommit=False:
            Specifies the autocommit settings.
            True will enable autocommit, False will disable it (default).
        - converter:
            Specifies a conversion dictionary, where keys are FIELD_TYPE values and values are conversion functions

    """
    if kwargs:
        if "pool_name" in kwargs:
            if not kwargs["pool_name"] in mariadb._CONNECTION_POOLS:
                pool= mariadb.ConnectionPool(**kwargs)
            else:
                pool= mariadb._CONNECTION_POOLS[kwargs["pool_name"]]
            c= pool.get_connection()
            return c

    connection= connectionclass(*args, **kwargs)
    if not isinstance(connection, mariadb.connections.Connection):
         raise mariadb.ProgrammingError("%s is not an instance of mariadb.Connection" % connection)
    return connection


Connection= connect

client_version_info= tuple(int(x, 10) for x in mariadb.mariadbapi_version.split('.'))
client_version= client_version_info[0] * 10000 + client_version_info[1] * 1000 + client_version_info[2]



#
# Copyright (C) 2020-2021 Georg Richter and MariaDB Corporation AB

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.

# You should have received a copy of the GNU Library General Public
# License along with this library; if not see <http://www.gnu.org/licenses>
# or write to the Free Software Foundation, Inc.,
# 51 Franklin St., Fifth Floor, Boston, MA 02110, USA
#

import mariadb, sys
import _thread, time

MAX_POOL_SIZE = 64
POOL_IDLE_TIMEOUT = 1800

class ConnectionPool(object):
    """
    Class defining a pool of database connections

    MariaDB Connector/Python supports simple connection pooling.
    A connection pool holds a number of open connections and handles
    thread safety when providing connections to threads.

    The size of a connection pool is configurable at creation time, 
    but cannot be changed afterwards. The maximum size of a connection
    pool is limited to 64 connections.

    Keyword Arguments:

        * pool_name (str) -- Name of connection pool

        * pool_size (int)=5 -- Size of pool. If not specified default value of 5 will be used. Maximum allowed number is 64.

        * pool_reset_connection (bool)=True -- Will reset the connection before returning it to the pool.  Default value is True.
    """

    def __init__(self, *args, **kwargs):
        """
        Creates a connetion pool class

        :param str pool_name:
            Name of connection pool

        :param int pool_size:
            Size of pool. If not specified default value of 5 will be used.
            Maximum allowed number is 64.

        :param bool pool_reset_connection:
            Will reset the connection before returning it to the pool.
            Default value is True.
        """
        self._connections = []
        self._pool_args = {}
        self._conn_args = {}
        self._lock_pool = _thread.RLock()
        self.__closed= 0

        key_words= ["pool_name", "pool_size", "pool_reset_connection"]

        # check if pool_name was provided
        if kwargs and "pool_name" in kwargs:

            # check if pool_name already exists
            if kwargs["pool_name"] in mariadb._CONNECTION_POOLS:
                raise mariadb.ProgrammingError("Pool '%s' already exists" \
                                               % kwargs["pool_name"])
        else:
            raise mariadb.ProgrammingError("No pool name specified")

        # save pool keyword arguments
        self._pool_args["name"]= kwargs.get("pool_name")
        self._pool_args["size"]= kwargs.get("pool_size", 5);
        self._pool_args["reset_connection"]= \
            kwargs.get("pool_reset_connection", True)

        # validate pool size (must be in range between 1 and MAX_POOL_SIZE)
        if not (0 < self._pool_args["size"] <= MAX_POOL_SIZE):
            raise mariadb.ProgrammError("Pool size must be in range of "\
                                        "1 and %s" % MAX_POOL_SIZE)

        # store pool and connection arguments
        self._conn_args= kwargs.copy()
        for key in key_words:
            if key in self._conn_args:
                del self._conn_args[key]

        if len(self._conn_args) > 0:
            with self._lock_pool:
                # fill connection pool
                for i in range(0, self._pool_args["size"]):
                    try:
                        connection= mariadb.Connection(**self._conn_args)
                    except:
                        # if an error occurred, close all connections and raise exception
                        for j in range(0, len(self._connections)):
                            try:
                                self._connections[j].close()
                            except:
                                # connect failed, so we are not interested in errors
                                # from close() method
                                pass
                            del self._connections[j]
                        raise
                    self.add_connection(connection)

        # store connection pool in _CONNECTION_POOLS 
        mariadb._CONNECTION_POOLS[self._pool_args["name"]]= self

    def __repr__(self):
        if (self.__closed):
            return "<mariadb.connectionPool.ConnectionPool object (closed) at %s>" % (hex(id(self)),)
        else:
            return "<mariadb.connectionPool.ConnectionPool object (name=%s) at %s>" % (self.pool_name, hex(id(self)))

    def add_connection(self, connection= None):
        """
        Adds a connection object to the connection pool.

        In case that the pool doesn’t have a free slot or is not configured
        a PoolError exception will be raised.
        """

        if not self._conn_args:
            raise mariadb.PoolError("Couldn't get configuration for pool %s" % \
                self._pool_args["name"])

        if connection is not None and \
            not isinstance(connection, mariadb.connections.Connection):
            raise mariadb.ProgrammingError("Passed parameter is not a connection object")

        if connection == None and len(self._conn_args) == 0:
            raise mariadb.PoolError("Can't get configuration for pool %s" % \
                self._pool_args["name"])

        if len(self._connections) >= self._pool_args["size"]:
            raise mariadb.PoolError("Can't add connection to pool %s: "
                "No free slot available (%s)." % (self._pool_args["name"], len(self._connections)))

        with self._lock_pool:
            if connection is None:
                connection= mariadb.Connection(**self._conn_args)
            
            connection._Connection__pool= self
            connection._Connection__in_use= 0
            connection._Connection__last_used= time.monotonic()
            self._connections.append(connection)
            

    def get_connection(self):
        """
        Returns a connection from the connection pool or raises a PoolError
        exception if a connection is not available.
        """
        
        now= time.monotonic()
        conn= None
        timediff= -1

        with self._lock_pool:
            for i in range(0, len(self._connections)):
                if not self._connections[i]._Connection__in_use:
                    try:
                        self._connections[i].ping()
                    except:
                        continue
                    t = now - self._connections[i]._Connection__last_used
                    if t > timediff:
                        conn= self._connections[i]
                        timediff = t

            if conn:
                conn._Connection__in_use= 1
        return conn

    def _close_connection(self, connection):
        """
        Returns connection to the pool. Internally used
        by connection object.
        """

        with self._lock_pool:
            if self._pool_args["reset_connection"]:
                connection.reset()
            connection._Connection__in_use= 0
            connection._Connection__last_used= time.monotonic()

    def set_config(self, **kwargs):
        """
        Sets the connection configuration for the connection pool.
        For valid connection arguments check the mariadb.connect() method.
        
        Note: This method doesn't create connections in the pool.
        To fill the pool one has to use add_connection() ḿethod.
        """

        self._conn_args= kwargs;

    def close(self):
        """Closes connection pool and all connections."""
        try:
            for c in self._connections:
                c._Connection__pool= None
                c.close()
        finally:
            self._connections= None
            del mariadb._CONNECTION_POOLS[self._pool_args["name"]]

    @property
    def pool_name(self):
        """Returns the name of the connection pool."""

        return self._pool_args["name"]

    @property
    def pool_size(self):
        """Returns the size of the connection pool."""

        return self._pool_args["size"]

    @property
    def max_size(self):
        "Returns the maximum size for connection pools."""

        return MAX_POOL_SIZE

    @property
    def connection_count(self):
        "Returns the number of connections in connection pool."""

        return len(self._connections)

    @property
    def pool_reset_connection(self):
        """
        If set to true, the connection will be reset on both client and server side\n
        after .close() method was called
        """
        return self._pool_args["reset_connection"]

    @pool_reset_connection.setter
    def pool_reset_connection(self, reset):
        self._pool_args["reset_connection"]= reset
